#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#if defined(__MSW__)
# include <windows.h>
#else
# include <unistd.h>
#endif
#include <gtk/gtk.h>

#include "../include/disk.h"

#include "guiutils.h"
#include "cdialog.h"
#include "pdialog.h"
#include "fprompt.h"
#include "fb.h"

#include "cfg.h"
#include "cfgfio.h"
#include "dspview.h"
#include "win.h"
#include "core.h"
#include "cfglist.h"
#include "config.h"


static void MainSignalCB(int s);
static gint MainTOCB(gpointer data);

win_struct *MainWinNew(core_struct *core);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	UNIX signal callback.
 */
static void MainSignalCB(int s)
{
	switch(s)
	{
	  case SIGINT:
	  case SIGTERM:
	  case SIGSEGV:
	    exit(1);
	    break;
	}
}

/*
 *	Main timeout callback.
 */
static gint MainTOCB(gpointer data)
{
	gint total_windows = 0;
	GList *glist;
	win_struct *win;
	core_struct *core = CORE(data);
	if(core == NULL)
	    return(FALSE);

	/* Manage all windows, count the number of windows that are
	 * mapped and delete any unmapped windows
	 */
	for(glist = core->win;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    win = WIN(glist->data);
	    if(win == NULL)
		continue;

	    if(WinIsMapped(win))
	    {
		total_windows++;
	    }
	    else
	    {
		WinDelete(win);
		glist->data = NULL;
		continue;
	    }
	}

	/* If no windows are mapped then break out of the main loop */
	if((total_windows == 0) || core->close_all_windows)
	{
	    gtk_main_quit();
	    return(FALSE);
	}

	return(TRUE);
}


/*
 *      Creates a new default Configuration List.
 */
static cfg_item_struct *MainCfgListNew(void)
{
	const cfg_item_struct src_cfg_list[] = CFG_LIST;
	cfg_item_struct *cfg_list = CFGItemListCopyList(src_cfg_list);
	gchar cwd[PATH_MAX]; 

	if(getcwd(cwd, sizeof(cwd)) != NULL)
	    cwd[sizeof(cwd) - 1] = '\0';
	else
	    *cwd = '\0';

	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_VERSION_MAJOR,
	    PROG_VERSION_MAJOR, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_VERSION_MINOR,
	    PROG_VERSION_MINOR, FALSE
	);                           
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_VERSION_RELEASE,
	    PROG_VERSION_RELEASE, FALSE
	);                             

	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_PLAY_ON_OPEN,
	    TRUE, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_AUTO_CHANGE_AUDIO,
	    TRUE, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_REPEATING,
	    FALSE, FALSE     
	);
	CFGItemListSetValueS(
	    cfg_list, CFG_PARM_RECORDER,
	    "localhost:9433", FALSE     
	);

	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_WIN_X,
	    0, FALSE
	);          
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_WIN_Y,
	    0, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_WIN_SHOW_DISPLAY,
	    TRUE, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_WIN_SHOW_VOLUME,
	    FALSE, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_WIN_SHOW_STATUS_BAR,
	    TRUE, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_WIN_DISPLAY_STYLE,
	    DSPVIEW_DISPLAY_STYLE_NOVA, FALSE
	);


	return(cfg_list);
}



/*
 *	Creates a new Win.
 */
win_struct *MainWinNew(core_struct *core)
{
	const gchar *recorder;
	GList *glist;
	const cfg_item_struct *cfg_list;
	win_struct *win = WinNew(core);
	if(win == NULL)
	    return(NULL);

	cfg_list = core->cfg_list;

	WinMap(win);
	if(core->iconic)
	    gtk_window_minimize(GTK_WINDOW(win->toplevel));


	/* Add pointer to the list */
	for(glist = core->win;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    if(glist->data == NULL)
	    {
		glist->data = win;
		return(win);
	    }
	}
	core->win = g_list_append(core->win, win);

	/* Connect to Y Server */
	recorder = CFGItemListGetValueS(cfg_list, CFG_PARM_RECORDER);
	WinSetBusy(win, TRUE);
	WinConnectionOpen(
	    win,
	    !STRISEMPTY(recorder) ? recorder : NULL
	);
	WinSetBusy(win, FALSE);

	return(win);
}


int main(int argc, char *argv[])
{
	gboolean initialized_gtk = FALSE;
	gchar cwd[PATH_MAX];
	GList *startup_file = NULL;
	gboolean	iconic = FALSE,
			play_on_open = TRUE,
			auto_change_audio = TRUE,
			repeating = FALSE,
			show_display = TRUE,
			show_volume = FALSE,
			show_status_bar = TRUE;
	dspview_display_style display_style = DSPVIEW_DISPLAY_STYLE_NOVA;
	const gchar	*title = NULL,
			*wm_name = NULL,
			*wm_class = NULL,
			*recorder = NULL;
	GdkGeometryFlags	geometry_flags = 0;
	GdkRectangle		*geometry = NULL;
	gint i;
	const gchar *arg;
	cfg_item_struct *cfg_list;
	win_struct *win;
	core_struct *core;

	/* Get current working directory */
	if(getcwd(cwd, sizeof(cwd)) != NULL)
	    cwd[sizeof(cwd) - 1] = '\0';
	else
	    *cwd = '\0';

	/* Set up time zone */
	tzset();

	/* Set up signal callbacks */
	signal(SIGINT, MainSignalCB);
	signal(SIGTERM, MainSignalCB);
	signal(SIGSEGV, MainSignalCB);
	signal(SIGSTOP, MainSignalCB);
	signal(SIGCONT, MainSignalCB);
	signal(SIGPIPE, MainSignalCB);


	/* Parse arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Help */
	    if(!g_strcasecmp(arg, "--help") ||
	       !g_strcasecmp(arg, "-help") ||
	       !g_strcasecmp(arg, "--h") ||
	       !g_strcasecmp(arg, "-h") ||
	       !g_strcasecmp(arg, "-?")
	    )
	    {
		g_print("%s", PROG_USAGE_MESG);
		return(0);
	    }
	    /* Version */
	    else if(!g_strcasecmp(arg, "--version") ||
		    !g_strcasecmp(arg, "-version")
	    )
	    {
		g_print("%s %s\n%s", PROG_NAME, PROG_VERSION, PROG_COPYRIGHT);
		return(0);
	    }
	    /* Show */
	    else if(g_strcasepfx(arg, "--show") ||
		    g_strcasepfx(arg, "-show")   
	    )
	    {
		while(*arg == '-')
		    arg++;
		while((*arg != '-') && (*arg != '\0'))
		    arg++;
		while(*arg == '-')    
		    arg++;

		if(!g_strcasecmp(arg, "display"))
		    show_display = TRUE;
		else if(!g_strcasecmp(arg, "volume"))
		    show_volume = TRUE;
		else if(!g_strcasecmp(arg, "status-bar"))
		    show_status_bar = TRUE;
	    }
	    /* Hide */
	    else if(g_strcasepfx(arg, "--hide") ||
		    g_strcasepfx(arg, "-hide")
	    )
	    {
		while(*arg == '-')    
		    arg++;   
		while((*arg != '-') && (*arg != '\0'))
		    arg++;
		while(*arg == '-')    
		    arg++;

		if(!g_strcasecmp(arg, "display"))
		    show_display = FALSE;
		else if(!g_strcasecmp(arg, "volume"))
		    show_volume = FALSE;
		else if(!g_strcasecmp(arg, "status-bar"))
		    show_status_bar = FALSE;
	    }
	    /* Play On Open */
	    else if(!g_strcasecmp(arg, "--play-on-open") ||
		    !g_strcasecmp(arg, "-play-on-open") ||
		    !g_strcasecmp(arg, "--p") ||
		    !g_strcasecmp(arg, "-p")
	    )
	    {
		play_on_open = TRUE;
	    }
	    /* No Play On Open */
	    else if(!g_strcasecmp(arg, "--no-play-on-open") ||
		    !g_strcasecmp(arg, "-no-play-on-open") ||
		    !g_strcasecmp(arg, "--np") ||
		    !g_strcasecmp(arg, "-np")
	    )
	    {
		play_on_open = FALSE;
	    }
	    /* Auto Change Audio */
	    else if(!g_strcasecmp(arg, "--auto-change-audio") ||
		    !g_strcasecmp(arg, "-auto-change-audio") ||
		    !g_strcasecmp(arg, "--m") ||
		    !g_strcasecmp(arg, "-m")
	    )
	    {
		auto_change_audio = TRUE;
	    }
	    /* No Change Audio */
	    else if(!g_strcasecmp(arg, "--no-change-audio") ||
		    !g_strcasecmp(arg, "-no-change-audio") ||
		    !g_strcasecmp(arg, "--nm") ||
		    !g_strcasecmp(arg, "-nm")   
	    )
	    {
		auto_change_audio = FALSE;
	    }
	    /* Repeating */
	    else if(!g_strcasecmp(arg, "--repeating") ||
		    !g_strcasecmp(arg, "-repeating") ||
		    !g_strcasecmp(arg, "--r") ||
		    !g_strcasecmp(arg, "-r")
	    )
	    {
		repeating = TRUE;
	    }
	    /* No Repeating */
	    else if(!g_strcasecmp(arg, "--no-repeating") ||
		    !g_strcasecmp(arg, "-no-repeating") ||
		    !g_strcasecmp(arg, "--nr") ||
		    !g_strcasecmp(arg, "-nr")
	    )
	    {
		repeating = FALSE;
	    }
	    /* Visualizer */
	    else if(!g_strcasecmp(arg, "--visualizer") ||
		    !g_strcasecmp(arg, "-visualizer") ||
		    !g_strcasecmp(arg, "--v") ||
		    !g_strcasecmp(arg, "-v")
	    )
	    {
		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    if(!g_strcasecmp(arg, "wave"))
			display_style = DSPVIEW_DISPLAY_STYLE_WAVE;
		    else if(!g_strcasecmp(arg, "sine-graph") ||
			    !g_strcasecmp(arg, "sgraph")
		    )
			display_style = DSPVIEW_DISPLAY_STYLE_SGRAPH;
		    else if(!g_strcasecmp(arg, "analyzer"))
			display_style = DSPVIEW_DISPLAY_STYLE_ANALYZER;
		    else if(!g_strcasecmp(arg, "lightening"))
			display_style = DSPVIEW_DISPLAY_STYLE_LIGHTENING;
		    else if(!g_strcasecmp(arg, "plasma"))
			display_style = DSPVIEW_DISPLAY_STYLE_PLASMA;
		    else if(!g_strcasecmp(arg, "nova"))
			display_style = DSPVIEW_DISPLAY_STYLE_NOVA;
		    else
			g_printerr(
			    "%s: Unsupported argument \"%s\".\n",
			    argv[i - 1], arg
			);
		}
		else
		{
		    g_printerr(
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }

	    /* Recorder */
	    else if(!g_strcasecmp(arg, "--recorder") ||
		    !g_strcasecmp(arg, "-recorder")
	    )
	    {
		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    recorder = arg;
		}
		else
		{
		    g_printerr(
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Title */
	    else if(!g_strcasecmp(arg, "--title") ||
		    !g_strcasecmp(arg, "-title")
	    )
	    {
		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    title = arg;
		}
		else
		{
		    g_printerr(
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* WM Name */
	    else if(!g_strcasecmp(arg, "--name") ||
		    !g_strcasecmp(arg, "-name")
	    )
	    {
		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    wm_name = arg;
		}
		else
		{
		    g_printerr(
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* WM Class */
	    else if(!g_strcasecmp(arg, "--class") ||
		    !g_strcasecmp(arg, "-class")
	    )
	    {
		i++;  
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    wm_class = arg;
		}
		else
		{
		    g_printerr(
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Geometry */
	    else if(!g_strcasecmp(arg, "--geometry") ||
		    !g_strcasecmp(arg, "-geometry")   
	    )
	    {
		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    gint x, y, width, height;
		    GdkRectangle *g = geometry;

		    if(g == NULL)
			geometry = g = (GdkRectangle *)g_malloc0(
			    sizeof(GdkRectangle)
			);

		    if(g != NULL)
		    {
			GdkGeometryFlags status = gdk_parse_geometry(
			    arg, &x, &y, &width, &height
			);
			if(status & GDK_GEOMETRY_X)
			    g->x = x;
			if(status & GDK_GEOMETRY_Y)
			    g->y = y;
			if(status & GDK_GEOMETRY_WIDTH)
			    g->width = width;
			if(status & GDK_GEOMETRY_HEIGHT)
			    g->height = height;
			geometry_flags = status;
		    }
		}
		else
		{
		    g_printerr(
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Iconic */
	    else if(!g_strcasecmp(arg, "--iconic") ||
		    !g_strcasecmp(arg, "-iconic") ||
		    !g_strcasecmp(arg, "--minimized") ||
		    !g_strcasecmp(arg, "-minimized")
	    )
	    {
		iconic = TRUE;
	    }

	    /* All else assume file name */
	    else if(*arg != '-')
	    {
		const gchar *path = arg;
		if(ISPATHABSOLUTE(path))
		    startup_file = g_list_append(
			startup_file,
			STRDUP(path)
		    );
		else
		    startup_file = g_list_append(
			startup_file,
			STRDUP(PrefixPaths(cwd, path))
		    );
	    }
	}


	/* Initialize GTK as needed */
	if(!initialized_gtk)
	{
	    if(!gtk_init_check(&argc, &argv))
	    {
		g_printerr(
"Unable to initialize GTK.\n"
		);
		return(1);
	    }

	    /* Initialize GDK RGB buffers */   
	    gdk_rgb_init();

	    initialized_gtk = TRUE;
	}

	/* Initialize dialogs */
	CDialogInit();
	PDialogInit();
	FPromptInit();
	FileBrowserInit();

	/* Create core */
	core = CORE(
	    g_malloc0(sizeof(core_struct))
	);
	if(core == NULL)
	    return(3);

	/* Configuration List */
	core->cfg_list = cfg_list = MainCfgListNew();

	core->busy_cur = gdk_cursor_new(GDK_WATCH);
	core->close_all_windows = FALSE;
	core->win = NULL;

	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_PLAY_ON_OPEN,
	    play_on_open, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_AUTO_CHANGE_AUDIO,
	    auto_change_audio, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_REPEATING,
	    repeating, FALSE
	);
	CFGItemListSetValueS(
	    cfg_list, CFG_PARM_RECORDER,
	    recorder, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_WIN_SHOW_DISPLAY,
	    show_display, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_WIN_SHOW_VOLUME,
	    show_volume, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_WIN_SHOW_STATUS_BAR,
	    show_status_bar, FALSE
	);
	CFGItemListSetValueI(
	    cfg_list, CFG_PARM_WIN_DISPLAY_STYLE,
	    (gint)display_style, FALSE
	);

	core->title = STRDUP(title);
	core->wm_name = STRDUP(wm_name);
	core->wm_class = STRDUP(wm_class);
	core->geometry = geometry;
	core->geometry_flags = geometry_flags;
	core->iconic = iconic;

	geometry = NULL;
	geometry_flags = 0;

	/* Create Wins */
	if(startup_file != NULL)
	{
	    const gchar *path;
	    GList *glist;

	    for(glist = startup_file;
		glist != NULL;
		glist = g_list_next(glist)
	    )
	    {
		path = (const gchar *)glist->data;

		win = MainWinNew(core);
		if(win != NULL)
		{
		    WinOpen(win, path);
		    if(win->play_on_open)
			WinPlay(win);
		}
	    }
	}
	else
	{
	    win = MainWinNew(core);
	}

	/* Delete list of startup files */
	g_list_foreach(startup_file, (GFunc)g_free, NULL);
	g_list_free(startup_file);
	startup_file = NULL;

	/* Set main timeout callback */
	gtk_timeout_add(
	    1000,		/* In milliseconds */
	    (GtkFunction)MainTOCB,
	    core
	);
	gtk_main();


	/* Delete core */
	g_list_foreach(core->win, (GFunc)WinDelete, NULL);
	g_list_free(core->win);
	core->win = NULL;

	GDK_CURSOR_DESTROY(core->busy_cur);
	core->busy_cur = NULL;

	g_free(core->title);
	g_free(core->wm_name);
	g_free(core->wm_class);
	g_free(core->geometry);

	/* Configuration List */
	CFGItemListDeleteList(core->cfg_list);

	g_free(core);


	/* Shutdown dialogs */
	FileBrowserShutdown();
	FPromptShutdown();
	PDialogShutdown();
	CDialogShutdown();

	return(0);
}
